# -*- coding: utf-8 -*-
"""
dlgpr.py
fdtd2d/fdtd3d対応
"""

import sys
import numpy as np
import torch
from torch import nn, optim
from torchvision import models
import matplotlib.pyplot as plt
#from torchinfo import summary

import utils, dl, fit
import plotutils

def main(argv):
    # FDTD計算結果ファイル名(誘電率/導電率分布と地中レーダー画像からなる)
    fdtd_bin = '../fdtd2d/fdtd.bin'
    #fdtd_bin = '../fdtd3d/fdtd.bin'

    # モデルファイル
    modelfile = 'dlgpr.pth'
    load_model = 0     # 通常0, 前回保存したmodelfileからrestartするときと推論時は1
    save_model = 1     # 通常1, 計算終了時にmodelfileを保存しないときと推論時は0

    # 計算パラメーター
    ndata = -1        # データ数, -1のときはすべてのデータ
    batch_size = 50    # バッチサイズ, 通常50-100程度
    num_epochs = 50    # エポック数(推論時は0)
    train_ratio = 0.8  # 訓練データの割合(通常0.8程度, 推論時は0)
    mixed = True and torch.cuda.is_available()  # 混合精度

    # device
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

    # 乱数初期化
    utils.torch_seed()

    # FDTD計算結果を読み込む(fdtd2dではNz=1)
    Ndata, Nx, Ny, Nz, NTx, NRx, Ntime, E, D = utils.read_data(fdtd_bin)
    #print(Ndata, Nx, Ny, Nz, NTx, NRx, Ntime); return

    # debug, データ図形確認
    #_plot2d_e(E); return
    #_plot3d_e(E); return
    #_plot2d_r(D); return

    # dataset
    train_set, test_set = dl.dataset(ndata, E, D, train_ratio)
    #image, label = next(iter(test_set)); print(image.shape, label.shape); return

    # dataloader
    train_loader, test_loader = dl.dataloader(train_set, test_set, batch_size=batch_size)
    #image, label = next(iter(test_loader)); print(image.shape, label.shape); return

    # 入力サイズ
    image, _ = next(iter(test_set))
    in_size = image.shape[0]

    # 出力サイズ
    out_size = Nx * Ny * Nz

    # ResNet
    # 重みなし
    #model = models.resnet18(weights=None)
    #model = models.resnet34(weights=None)
    #model = models.resnet50(weights=None)
    # 重みあり
    #model = models.resnet18(weights=models.ResNet18_Weights.DEFAULT)
    model = models.resnet34(weights=models.ResNet34_Weights.DEFAULT)
    #model = models.resnet50(weights=models.ResNet50_Weights.DEFAULT)
    # model付け替え
    model.conv1 = nn.Conv2d(in_size, 64, kernel_size=7, stride=2, padding=3, bias=False)
    model.fc = nn.Linear(model.fc.in_features, out_size)
    
    # debug
    #summary(model, (50, 3, 112, 112)); return
    #print(model); return

    # restart（前回と同じmodelのとき有効）
    if load_model == 1:
        utils.load_model(model, modelfile, device)

    # 損失関数
    criterion = nn.MSELoss()

    # 最適化関数
    #optimizer = optim.Adam(model.parameters())
    optimizer = optim.NAdam(model.parameters())

    # 計算条件表示
    image, _ = next(iter(test_set))
    print('train=%d test=%d %dx%dx%d %dx%dx%d=%d epochs=%d'
        % (len(train_set), len(test_set), image.shape[0], image.shape[1], image.shape[2], Nx, Ny, Nz, Nx * Ny * Nz, num_epochs))

    # 学習実行(計算時間の主要部)
    preds, history = fit.fit(model, device, num_epochs, train_loader, test_loader, optimizer, criterion, mixed,
        modelfile if (save_model == 1) else None, step=5)

    # 計算結果をファイルに保存する
    datafile = 'dlgpr.npz'
    utils.save_data(datafile, Nx, Ny, Nz, test_set, preds)

    # debug, loss履歴を図形表示する
    _plot2d_history(history)

    # debug, loss履歴をloss.logに出力する
    _output_history(history, 'loss.log')

# debug, 誘電率/導電率分布を2D図形表示する
def _plot2d_e(E):
    # E[ndata, nx, ny, nz], f:3D配列
    #print(E.shape, np.min(E), np.max(E))
    assert E.ndim == 4
    if E.shape[3] == 1:
        # fdtd2d:
        f = E[:, :, :, 0]
        figtitle = '誘電率/導電率分布(fdtd2d)'
    else:
        # fdtd3d
        n = 0  # データ番号を適当に与える
        n = max(0, min(E.shape[0] - 1, n))
        f = E[n, :, :, :].transpose((2, 0, 1))  # ZXYに変更する
        figtitle = '誘電率/導電率分布2D(fdtd3d %d/%d)' % (n + 1, E.shape[0])
    sta = [str(i + 1) for i in range(f.shape[0])]

    # 図形出力
    plotutils.contour(figtitle, (8, 6), 6, 7, 1, f, np.min(E), np.max(E), sta, cmap='rainbow')

# debug, 誘電率/導電率分布を3D図形表示する
def _plot3d_e(E):
    # E[ndata, nx, ny, nz]
    if E.shape[3] == 1:  # fdtd3d限定
        return
    n = 0  # データ番号を適当に与える
    n = max(0, min(E.shape[0] - 1, n))
    figtitle = '誘電率/導電率分布3D(fdtd3d %d/%d)' % (n + 1, E.shape[0])

    # 図形出力
    plotutils.plot3d_surface(figtitle, E[n], figsize=(5, 7), stride=2, zfactor=5)

# debug, レーダー画像を2D図形表示する
def _plot2d_r(D):
    # D[ndata, ntx, nrx, ntime]
    #print(D.shape, np.min(D), np.max(D))
    assert D.ndim == 4
    if D.shape[2] == 1:
        # 1受信点
        f = D[:, :, 0, :]  # 3D配列
        sta = [str(i + 1) for i in range(D.shape[0])]
    else:
        # 複数受信点: 受信点配列をデータ配列に合体する(my=1,mstep=1とする)
        f = np.zeros((D.shape[0] * D.shape[2], D.shape[1], D.shape[3]))  # 3D配列
        sta = []
        idx = 0
        for idata in range(D.shape[0]):
            for irx in range(D.shape[2]):
                f[idx, :, :] = D[idata, :, irx, :]
                sta.append('%d-%d' % (idata + 1, irx + 1))
                idx += 1

    figtitle = 'レーダー画像'
    figsize = (9, 6)  # 横,縦(inch)
    dmin, dmax = np.min(D), np.max(D)
    da = max(-dmin, dmax)
    dmin, dmax = -da, da  # 最小値=-最大値

    # 図形出力
    plotutils.contour(figtitle, figsize, 5, 1, 1, f, dmin, dmax, sta, flipud=True, cmap='RdBu_r')

# debug, loss履歴を図形表示する
def _plot2d_history(history):
    if len(history) < 2:
        return
    plt.figure('loss', figsize=(6, 4), layout='tight')
    plt.plot(history)
    plt.xlim(0, len(history) - 1)
    plt.ylim(0, 0.3)  # maxは適当に編集する
    plt.grid(True)
    plt.xlabel('epoch')
    plt.ylabel('loss')
    plt.show()

# debug, loss履歴をファイルに出力する
def _output_history(history, fn):
    fp = open(fn, 'wt', encoding='utf-8')
    for i, loss in enumerate(history):
        fp.write('%d %.5f\n' % (i, loss))
    fp.close()

if __name__ == "__main__":
    main(sys.argv)
