# -*- coding: utf-8 -*-
"""
plot.py
計算結果を図形表示する
"""

import os, datetime
import numpy as np

import utils, plotutils

def main():
    
    # テストデータのラベルと予測値をファイルから読み込む
    datafile = 'dlgpr.npz'
    Nx, Ny, Nz, labels, preds = utils.load_data(datafile)
    #print(Nx, Ny, Nz, labels.shape, preds.shape)

    # 1D->4D配列
    labels = labels.reshape(-1, Nx, Ny, Nz)
    preds  = preds.reshape(-1, Nx, Ny, Nz)
    #print(labels.shape, preds.shape)
    assert labels.shape == preds.shape

    # ファイル更新日時
    dt = datetime.datetime.fromtimestamp(os.path.getmtime(datafile)).strftime('%Y/%m/%d %H:%M:%S')

    # 最小・最大(必要なら手動で設定する)
    vmin = min(np.min(labels), np.min(preds))
    vmax = max(np.max(labels), np.max(preds))
    vmin = max(0, vmin)  # 最小値は0以上
    #print(vmin, vmax)

    # 等高線図
    figsize = (8, 6)
    mx = 5
    my = 10
    if labels.shape[3] == 1:
        # 2Dモデル
        _plot2d_data2d(figsize, mx, my, labels, preds, vmin, vmax, dt)
    else:
        # 3Dモデル
        _plot2d_data3d(figsize, mx, my, labels, preds, vmin, vmax, dt)
        _plot3d_data3d(preds, dt)

# 2Dモデル
def _plot2d_data2d(figsize, mx, my, labels, preds, vmin, vmax, dt):
    figtitle = 'fdtd2d label-pred %s' % dt
    p1 = labels[:,:,:,0]  # 4D->3D
    p2 = preds[:,:,:,0]
    mstep = int(np.ceil(p1.shape[0] / (mx * my)))  # step自動計算

    # 平均, 誤差
    ave = labels.mean(axis=(1,2,3))
    dif = np.abs(labels - preds).mean(axis=(1,2,3))
    str_ave = ['%d %.4f' % (i + 1, d) for i, d in enumerate(ave)]
    str_dif = ['%d %.4f' % (i + 1, d) for i, d in enumerate(dif)]

    # 図形出力
    plotutils.contour2(figtitle, figsize, mx, my, mstep, p1, p2, vmin, vmax, str_ave, str_dif)

# 3Dモデル
def _plot2d_data3d(figsize, mx, my, labels, preds, vmin, vmax, dt):
    n = 100  # データ番号: 適当に代入する
    n = max(0, min(n, labels.shape[0] - 1))

    figtitle = 'fdtd3d label-pred data#=%d %s' % (n + 1, dt)
    p1 = labels[n].transpose(2, 0, 1)  # XYZをZXYの順に変換する
    p2 = preds[n].transpose(2, 0, 1)
    mstep = int(np.ceil(p1.shape[0] / (mx * my)))  # step自動計算

    # 平均, 誤差
    ave = p1.mean(axis=(1,2))
    dif = np.abs(p1 - p2).mean(axis=(1,2))
    str_ave = ['%d %.4f' % (i + 1, d) for i, d in enumerate(ave)]
    str_dif = ['%d %.4f' % (i + 1, d) for i, d in enumerate(dif)]

    # 図形出力
    plotutils.contour2(figtitle, figsize, mx, my, mstep, p1, p2, vmin, vmax, str_ave, str_dif)

# 3Dモデルを3D図形表示する
def _plot3d_data3d(preds, dt):
    n = 100  # データ番号: 適当に代入する
    n = max(0, min(n, preds.shape[0] - 1))
    figtitle = 'fdtd3d pred data#=%d %s' % (n + 1, dt)

    # 図形出力
    plotutils.plot3d_surface(figtitle, preds[n], figsize=(5, 7), stride=2, zfactor=5)

if __name__ == "__main__":
    main()
