# -*- coding: utf-8 -*-
"""
utils.py
"""

import os, sys
import numpy as np
import torch

# datasetを読み込む
def read_data(fn):

    if not os.path.isfile(fn):
        print('file %s not found' % fn)
        sys.exit()

    print('Load %s' % fn)

    # データ数
    offset = 0
    count = 7
    idata = np.fromfile(fn, dtype='i4', count=count, offset=offset)
    Ndata, Nx, Ny, Nz, NTx, NRx, Ntime = idata
    offset += count * 4  # 4:i4
    #print(Ndata, Nx, Ny, Nz, NTx, NRx, Ntime)

    # 誘電率/導電率とレーダー画像の最小・最大
    count = 4
    fdata = np.fromfile(fn, dtype='f4', count=count, offset=offset)
    emin, emax, dmin, dmax = fdata
    offset += count * 4  # 4:f4
    #print(emin, emax, dmin, dmax)

    # 誘電率/導電率(u1->f4)
    count = Ndata * Nx * Ny * Nz
    e = np.fromfile(fn, dtype='u1', count=count, offset=offset)
    e = e.astype(np.float32) / 255
    e = emin + (emax - emin) * e
    E = e.reshape(-1, Nx, Ny, Nz)  # 4D化
    offset += count * 1  # 1:u1
    #print(E.shape)
    #print(np.min(E), np.max(E))

    # レーダー画像(u1->f4)
    count = Ndata * NTx * NRx * Ntime
    d = np.fromfile(fn, dtype='u1', count=count, offset=offset)
    d = d.astype(np.float32) / 255
    d = dmin + (dmax - dmin) * d
    D = d.reshape(-1, NTx, NRx, Ntime)  # 4D化
    #print(D.shape)
    #print(np.min(D), np.max(D))

    return Ndata, Nx, Ny, Nz, NTx, NRx, Ntime, E, D

# 乱数初期化
def torch_seed(seed = 123):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.use_deterministic_algorithms = True

# モデルファイルを読み込む
def load_model(model, modelfile, device):
    if not os.path.isfile(modelfile):
        print('file %s not found' % modelfile)
        sys.exit()
    model.load_state_dict(torch.load(modelfile, map_location=device, weights_only=True))
    print("Load %s" % modelfile)

# テストデータのラベルと予測値をファイルに保存する
def save_data(fn, Nx, Ny, Nz, test_set, preds):
    labels = []
    for _, label in test_set:
        labels.append(label.numpy())
    labels = np.array(labels).reshape(-1)

    np.savez(fn, Nx, Ny, Nz, labels, preds)
    print('Save %s' % fn)

# テストデータのラベルと予測値をファイルから読み込む
def load_data(fn):
    d = np.load(fn)
    Nx     = d[d.files[0]]
    Ny     = d[d.files[1]]
    Nz     = d[d.files[2]]
    labels = d[d.files[3]]
    preds  = d[d.files[4]]
    print('Load %s' % fn)

    return Nx, Ny, Nz, labels, preds

"""
# binデータをnpz形式で保存する
def save_bin(fn, Ndata, Nx, Ny, Nz, NTx, NRx, Ntime, E, D):
    np.savez(fn, Ndata, Nx, Ny, Nz, NTx, NRx, Ntime, E, D)
    print('Save %s' % fn)

# npz形式のbinデータを読み込む
def load_bin(fn):
    d = np.load(fn)
    Ndata = d[d.files[0]]
    Nx    = d[d.files[1]]
    Ny    = d[d.files[2]]
    Nz    = d[d.files[3]]
    NTx   = d[d.files[4]]
    NRx   = d[d.files[5]]
    Ntime = d[d.files[6]]
    E     = d[d.files[7]]
    D     = d[d.files[8]]
    print('Load %s' % fn)

    return Ndata, Nx, Ny, Nz, NTx, NRx, Ntime, E, D
"""
