/*
geometry.c
誘電率、導電率分布
セル中心
任意個数長方形
ランダムに変化: 縦の長さ, 横の長さ, 縦の位置, 横の位置, 誘電率, 導電率
境界面では両側の平均をとる
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>

extern double urand(int, int64_t *);
extern void getspan(int, int, int, int, int64_t *, int *, int *);

void geometry(int Ndata, int Nx, int Ny, int Ly, int Nrect, int Nsmot, int64_t Seed,
	const int Lrect[], const float Eps[], const float Sig[],
	float Eps0, float Sig0, const float Eps1[], const float Sig1[], float pLy1, const int iLy1[],
	float ***Epsr, float ***Sigm, int **rect)
{
	assert((Nx > 0) && (Ny > 0) && (Ly > 0) && (Nrect >= 0) && (Nsmot >= 0));
	assert((Ly < Ny) && (iLy1[0] < Ly) && (iLy1[1] < Ly));
	const int nout = Lrect[2];
	float r;

	// データに関するループ
	for (int idata = 0; idata < Ndata; idata++) {
		// pointer
		float **e = Epsr[idata];
		float **s = Sigm[idata];

		// 背景媒質, 地下境界面の有無で分ける
		r = (float)urand(3, &Seed);
		if (r < pLy1) {
			// (1) 地下境界面あり
			// 地下境界面のY座標
			r = (float)urand(3, &Seed);
			const int ly = iLy1[0] + (int)(r * (iLy1[1] - iLy1[0]));
			// 地下境界面の下の比誘電率と導電率
			r = (float)urand(3, &Seed);
			const float eps1a = Eps1[0] + (Eps1[1] - Eps1[0]) * r;
			r = (float)urand(3, &Seed);
			const float sig1a = Sig1[0] + (Sig1[1] - Sig1[0]) * r;
			// 地下境界面の上の比誘電率と導電率
			r = (float)urand(3, &Seed);
			const float eps1b = Eps1[0] + (Eps1[1] - Eps1[0]) * r;
			r = (float)urand(3, &Seed);
			const float sig1b = Sig1[0] + (Sig1[1] - Sig1[0]) * r;

			// 比誘電率と導電率を代入する
			for (int i = 0; i < Nx; i++) {
				// 地下境界面の下
				for (int j = 0; j < ly; j++) {
					e[i][j] = eps1a;
					s[i][j] = sig1a;
				}
				// 地下境界面
				e[i][ly] = (eps1a + eps1b) / 2;
				s[i][ly] = (sig1a + sig1b) / 2;
				// 地下境界面の上
				for (int j = ly + 1; j < Ly; j++) {
					e[i][j] = eps1b;
					s[i][j] = sig1b;
				}
				// 地面
				e[i][Ly] = (eps1b + Eps0) / 2;
				s[i][Ly] = (sig1b + Sig0) / 2;
				// 地上
				for (int j = Ly + 1; j < Ny; j++) {
					e[i][j] = Eps0;
					s[i][j] = Sig0;
				}
			}
		}
		else {
			// (2) 地下境界面なし
			// 地下の比誘電率と導電率
			r = (float)urand(3, &Seed);
			const float eps1 = Eps1[0] + (Eps1[1] - Eps1[0]) * r;
			r = (float)urand(3, &Seed);
			const float sig1 = Sig1[0] + (Sig1[1] - Sig1[0]) * r;

			// 比誘電率と導電率を代入する
			for (int i = 0; i < Nx; i++) {
				// 地下
				for (int j = 0; j < Ly; j++) {
					e[i][j] = eps1;
					s[i][j] = sig1;
				}
				// 地面
				e[i][Ly] = (Eps0 + eps1) / 2;
				s[i][Ly] = (Sig0 + sig1) / 2;
				// 地上
				for (int j = Ly + 1; j < Ny; j++) {
					e[i][j] = Eps0;
					s[i][j] = Sig0;
				}
			}
		}

		// 長方形, 誘電率/導電率代入
		for (int irect = 0; irect < Nrect; irect++) {
			const float r1 = (float)urand(3, &Seed);
			const float r2 = (float)urand(3, &Seed);
			const float ep = Eps[0] + (Eps[1] - Eps[0]) * r1;
			const float si = Sig[0] + (Sig[1] - Sig[0]) * r2;
			int i0, i1, j0, j1;
			getspan(nout, Nx - nout, Lrect[0], Lrect[1], &Seed, &i0, &i1);
			getspan(nout, Ly - nout, Lrect[0], Lrect[1], &Seed, &j0, &j1);
			//printf("%d %d %d %d\n", idata, irect, i0, i1);
			//printf("%d %d %d %d\n", idata, irect, j0, j1);
			for (int i = i0; i < i1; i++) {
			for (int j = j0; j < j1; j++) {
				e[i][j] = ep;
				s[i][j] = si;
			}
			}
			// debug, 長方形（図形出力用）
			rect[idata][(4 * irect) + 0] = i0;
			rect[idata][(4 * irect) + 1] = j0;
			rect[idata][(4 * irect) + 2] = i1;
			rect[idata][(4 * irect) + 3] = j1;
		}

		// 内部領域平滑化(5点平均, 複数回可)
		for (int m = 0; m < Nsmot; m++) {
			for (int i = nout; i < Nx - nout; i++) {
			for (int j = nout; j < Ly - nout; j++) {
				e[i][j] = (
				e[i][j] + e[i - 1][j    ] + e[i + 1][j    ]
				        + e[i    ][j - 1] + e[i    ][j + 1]) / 5;
				s[i][j] = (
				s[i][j] + s[i - 1][j    ] + s[i + 1][j    ]
				        + s[i    ][j - 1] + s[i    ][j + 1]) / 5;
			}
			}
		}
	}
}
