/*
misc.c
*/

#include <stdio.h>
#include <math.h>
#include <assert.h>

static void yee2cell(int, int, int, int, int *, int *, int *, int *);

extern float C;

// メッシュ
void make_mesh(int Nx, int Ny, int Ly, float Dx, float Dy,
	float *Xn, float *Yn, float *Dt,
	float *dXn, float *dYn, float *dXc, float *dYc)
{
	// 格子点
	for (int i = 0; i <= Nx; i++) {
		Xn[i] = (i - Nx / 2.0f) * Dx;
	}
	for (int j = 0; j <= Ny; j++) {
		Yn[j] = (j - Ly) * Dy;  // 地表面がY=0
	}
	//for (int i = 0; i <= Nx; i++) printf("%d %f\n", i, Xn[i]);
	//for (int j = 0; j <= Ny; j++) printf("%d %f\n", j, Yn[j]);

	// タイムステップ
	*Dt = (float)(1 / sqrt(1 / (Dx * Dx) + 1 / (Dy * Dy)) / C);
	//printf("%f %f %e\n", Dx, Dy, *Dt);

	// C * Dt
	const float cdt = C * (*Dt);

	// 格子点に関する因子
	for (int i = 1; i < Nx; i++) {
		dXn[i] = cdt / ((Xn[i + 1] - Xn[i - 1]) / 2);
	}
	for (int j = 1; j < Ny; j++) {
		dYn[j] = cdt / ((Yn[j + 1] - Yn[j - 1]) / 2);
	}
	dXn[0]  = cdt / (Xn[1]  - Xn[0]);
	dYn[0]  = cdt / (Yn[1]  - Yn[0]);
	dXn[Nx] = cdt / (Xn[Nx] - Xn[Nx - 1]);
	dYn[Ny] = cdt / (Yn[Ny] - Yn[Ny - 1]);

	// セル中心に関する因子
	for (int i = 0; i < Nx; i++) {
		dXc[i] = cdt / (Xn[i + 1] - Xn[i]);
	}
	for (int j = 0; j < Ny; j++) {
		dYc[j] = cdt / (Yn[j + 1] - Yn[j]);
	}
	//for (int i = 0; i <= Nx; i++) printf("%d %f\n", i, dXn[i]);
	//for (int j = 0; j <= Ny; j++) printf("%d %f\n", j, dYn[j]);
	//for (int i = 0; i <  Nx; i++) printf("%d %f\n", i, dXc[i]);
	//for (int j = 0; j <  Ny; j++) printf("%d %f\n", j, dYc[j]);
}


// 送信点座標
void tx(int NTx, int iTx0, int DTx, int Ly, int Lyant, int *iTx, int *jTx)
{
	for (int tx = 0; tx < NTx; tx++) {
		iTx[tx] = iTx0 + (tx * DTx);
		jTx[tx] = Ly + Lyant;
		//printf("%d %d %d\n", tx, iTx[tx], jTx[tx]);
	}
}


// Yee格子物性値
void setupYee(int Nx, int Ny, int Labc, float **epsr, float **sigm, float ***PEz)
{
	for (int i = -Labc; i < Nx + 1 + Labc; i++) {
	for (int j = -Labc; j < Ny + 1 + Labc; j++) {
		float eps, sig;
		if ((i > 0) && (i < Nx) && (j > 0) && (j < Ny)) {
			// 内部:4点の平均
			eps = (epsr[i - 1][j - 1] + epsr[i - 1][j] + epsr[i][j - 1] + epsr[i][j]) / 4;
			sig = (sigm[i - 1][j - 1] + sigm[i - 1][j] + sigm[i][j - 1] + sigm[i][j]) / 4;
		}
		else {
			// 外部:1点または2点の平均
			int i1, i2, j1, j2;
			yee2cell(Nx, Ny, i, j, &i1, &i2, &j1, &j2);
			eps = (epsr[i1][j1] + epsr[i2][j2]) / 2;
			sig = (sigm[i1][j1] + sigm[i2][j2]) / 2;
		}
		PEz[i + Labc][j + Labc][0] = 1 / eps;  // PMLの計算時間短縮のために逆数とする
		PEz[i + Labc][j + Labc][1] = sig;
	}
	}
}


// FDTD係数
void setupFactor(int Nx, int Ny, int Labc, float Dt, float ***PEz, float **C1Ez, float **C2Ez)
{
	const float pi = (float)(4 * atan(1));
	const float eta = C * 4 * pi * 1e-7f;

	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 1; j++) {
		const float eps = 1 / PEz[i + Labc][j + Labc][0];  // PEzは比誘電率の逆数
		const float sig =     PEz[i + Labc][j + Labc][1];
		C2Ez[i][j] = 1 / (eps + (sig * eta * C * Dt));
		C1Ez[i][j] = C2Ez[i][j] * eps;
		//printf("%d %d %f %f\n", i, j, C1Ez[i][j], C2Ez[i][j]);
	}
	}
}


// 外部または境界の節点番号(i, j)に一番近い内部の2つのセル番号(i1, j1),(i2, j2)を求める
static void yee2cell(int nx, int ny, int i, int j, int *i1, int *i2, int *j1, int *j2)
{
	*i1 = *i2 = i;
	*j1 = *j2 = j;
	// 4隅
	if      ((i <=  0) & (j <=  0)) {
		*i1 = *i2 = 0;
		*j1 = *j2 = 0;
	}
	else if ((i <=  0) & (j >= ny)) {
		*i1 = *i2 = 0;
		*j1 = *j2 = ny - 1;
	}
	else if ((i >= nx) & (j <=  0)) {
		*i1 = *i2 = nx - 1;
		*j1 = *j2 = 0;
	}
	else if ((i >= nx) & (j >= ny)) {
		*i1 = *i2 = nx - 1;
		*j1 = *j2 = ny - 1;
	}
	// 4辺
	else if (i <=  0) {
		*i1 = *i2 = 0;
		*j1 = j - 1;
		*j2 = j;
	}
	else if (i >= nx) {
		*i1 = *i2 = nx - 1;
		*j1 = j - 1;
		*j2 = j;
	}
	else if (j <=  0) {
		*i1 = i - 1;
		*i2 = i;
		*j1 = *j2 = 0;
	}
	else if (j >= ny) {
		*i1 = i - 1;
		*i2 = i;
		*j1 = *j2 = ny - 1;
	}
	assert((*i1 >= 0) && (*i1 < nx));
	assert((*i2 >= 0) && (*i2 < nx));
	assert((*j1 >= 0) && (*j1 < ny));
	assert((*j2 >= 0) && (*j2 < ny));
}


// メモリーサイズ[MB]
int memory_size(int nthread, int Ndata, int Nx, int Ny, int NTx, int Niter, int Labc, int fig2)
{
	size_t mem = 0;

	mem += Ndata * Nx * Ny * 2 * sizeof(float);       // Epsr, Sigm
	mem += Ndata * NTx * Niter * sizeof(float);       // Er
	mem += nthread * Nx * Ny * 2 * sizeof(float);     // C1Ez, C2Ez
	mem += nthread * (Nx + 2 * Labc) * (Ny + 2 * Labc) * 3 * sizeof(float);  // Ez, Hx, Hy
	mem += nthread * (Nx + 2 * Labc) * (Ny + 2 * Labc) * (Labc == 0 ? 4 : 3) * sizeof(float);  // ABC
	mem += (fig2 > 0) * Nx * Ny * 3 * sizeof(float);  // Ez_r, Ez_i, Ez_a

	return (int)(mem * 1e-6) + 1;
}
