/*
solve.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "alloc.h"

static void initfield(int, int, int, int, float **, float **, float **, float *, float ***, float ***, float **, float **, int, float **, float **);
static void average(int, int, int, float **, float **, float **, float []);
static void dft(int, int, int, float **, float, float, float **, float **);

extern void updateEz(int, int, int, float **, float **, float **, float *, float *, float **, float **);
extern void updateHx(int, int, int, float **, float **, float *);
extern void updateHy(int, int, int, float **, float **, float *);
extern float vfeed(float, float, float, int);
extern void murEz(int, int, float **, float ***, float ***);
extern void pmlEz(int, int, int, float **, float **, float **, float **, float **, int **, float ***, float *, float *, float *, float *);
extern void pmlHx(int, int, int, float **, float **, int **, float *, float *);
extern void pmlHy(int, int, int, float **, float **, int **, float *, float *);

extern float C;

void solve(
	int Niter, int Nx, int Ny, float Freq, float Dt, int Labc, int WF, int Nant, float Dy, int ilog,
	int iTx, int jTx, int LTxRx,
	float **Ez, float **Hx, float **Hy,
	float *dXn, float *dYn, float *dXc, float *dYc,
	float ***PEz, float **C1Ez, float **C2Ez,
	float ***Ezmurx, float ***Ezmury,
	float **Ezpmlx, float **Ezpmly, int **bEzpml, int **bHxpml, int **bHypml,
	float *fEzpmlx, float *fEzpmly, float *fHxpml, float *fHypml,
	float *dXnpml, float *dXcpml, float *dYnpml, float *dYcpml,
	float *Er, int iout, float *Fcos, float *Fsin, float **ez_r, float **ez_i)
{
	//printf("%d %d %d %e %e %d %d %d\n", Niter, Nx, Ny, Freq, Dt, Labc, iTx, jTx); fflush(stdout);

	// 電磁界初期化
	initfield(Nx, Ny, Labc, Niter, Ez, Hx, Hy, Er, Ezmurx, Ezmury, Ezpmlx, Ezpmly, iout, ez_r, ez_i);

	// FDTD反復計算
	float t = 0;
	for (int iter = 0; iter <= Niter; iter++) {
		// Ez更新
		updateEz(Nx, Ny, Labc, Ez, Hx, Hy, dXn, dYn, C1Ez, C2Ez);

		// 給電点
		Ez[iTx + Labc][jTx + Labc] = vfeed(t, 0, Freq, WF);
		//printf("%d %e %e\n", iter, t, Ez[iTx + Labc][jTx + Labc]);
		if (Nant == 2) {
			// アレーアンテナ
			const int ya2 = (int)(C / (4 * Freq * Dy) + 0.5);  // 1/4波長セル数
			const float delay2 = -1 / (4 * Freq);  // 遅延時間
			Ez[iTx + Labc][jTx + ya2 + Labc] = vfeed(t, delay2, Freq, WF);
		}

		// ABC
		if (Labc > 0) {
			pmlEz(Nx, Ny, Labc, Ez, Hx, Hy, Ezpmlx, Ezpmly, bEzpml, PEz, fEzpmlx, fEzpmly, dXnpml, dYnpml);
		}
		else {
			murEz(Nx, Ny, Ez, Ezmurx, Ezmury);
		}

		// Hx/Hy更新
		updateHx(Nx, Ny, Labc, Ez, Hx, dYc);
		updateHy(Nx, Ny, Labc, Ez, Hy, dXc);

		// ABC
		if (Labc > 0) {
			pmlHx(Nx, Ny, Labc, Ez, Hx, bHxpml, fHxpml, dYcpml);
			pmlHy(Nx, Ny, Labc, Ez, Hy, bHypml, fHypml, dXcpml);
		}

		// 受信波形（送信点の+X側）
		Er[iter] = Ez[iTx + LTxRx + Labc][jTx + Labc];
		//printf("%d %e %e\n", iter, t, Er[iter]);

		// DFT, debug
		if (iout) {
			dft(Nx, Ny, Labc, Ez, Fcos[iter], Fsin[iter], ez_r, ez_i);
		}

		// 時間ステップ
		t += Dt;

		// debug, 平均電磁界出力
		if (ilog && (iter % 20 == 0)) {
			float fsum[2];
			average(Nx, Ny, Labc, Ez, Hx, Hy, fsum);
			printf("%5d %.6f %.6f\n", iter, fsum[0], fsum[1]);
			fflush(stdout);
		}
	}
}


// 電磁界初期化
static void initfield(int Nx, int Ny, int Labc, int Niter,
	float **Ez, float **Hx, float **Hy,
	float *Er,
	float ***Ezmurx, float ***Ezmury,
	float **Ezpmlx, float **Ezpmly,
	int iout, float **ez_r, float **ez_i)
{
	assert((Nx > 0) && (Ny > 0) && (Labc >= 0) && (Niter > 0));

	for (int i = - Labc; i < Nx + 1 + Labc; i++) {
		memset(Ez[i + Labc], 0, (Ny + 1 + 2 * Labc) * sizeof(float));
	}
	for (int i = - Labc; i < Nx + 1 + Labc; i++) {
		memset(Hx[i + Labc], 0, (Ny + 0 + 2 * Labc) * sizeof(float));
	}
	for (int i = - Labc; i < Nx + 0 + Labc; i++) {
		memset(Hy[i + Labc], 0, (Ny + 1 + 2 * Labc) * sizeof(float));
	}

	// 波形
	memset(Er, 0, Niter * sizeof(float));

	// ABC
	if (Labc == 0) {
		for (int i = 0; i < Nx + 1; i++) {
		for (int j = 0; j < Ny + 1; j++) {
			memset(Ezmurx[i][j], 0, 2 * sizeof(float));
			memset(Ezmury[i][j], 0, 2 * sizeof(float));
		}
		}
	}
	else {
		for (int i = - Labc; i < Nx + 1 + Labc; i++) {
			memset(Ezpmlx[i + Labc], 0, (Ny + 1 + 2 * Labc) * sizeof(float));
			memset(Ezpmly[i + Labc], 0, (Ny + 1 + 2 * Labc) * sizeof(float));
		}
	}

	// debug
	if (iout) {
		for (int i = 0; i < Nx + 1; i++) {
			memset(ez_r[i], 0, (Ny + 1) * sizeof(float));
			memset(ez_i[i], 0, (Ny + 1) * sizeof(float));
		}
	}
}


// 平均電磁界
static void average(int Nx, int Ny, int Labc, float **Ez, float **Hx, float **Hy, float fsum[])
{
	// <E>
	float sume = 0;
	int nume = 0;
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 1; j++) {
		sume += (float)fabs(Ez[i + Labc][j + Labc]);
		nume++;
	}
	}
	fsum[0] = sume/ nume;

	// <H>
	float sumh = 0;
	int numh = 0;
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 0; j++) {
		sumh += (float)fabs(Hx[i + Labc][j + Labc]);
		numh++;
	}
	}
	for (int i = 0; i < Nx + 0; i++) {
	for (int j = 0; j < Ny + 1; j++) {
		sumh += (float)fabs(Hy[i + Labc][j + Labc]);
		numh++;
	}
	}
	fsum[1] = sumh / numh;
}


// DFT
static void dft(int Nx, int Ny, int Labc, float **Ez, float fcos, float fsin, float **Ez_r, float **Ez_i)
{
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 1; j++) {
		Ez_r[i][j] += fcos * Ez[i + Labc][j + Labc];
		Ez_i[i][j] -= fsin * Ez[i + Labc][j + Labc];
	}
	}
}
