/*
utils.c
fdtd2d/fdtd3d共通
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <assert.h>

extern double urand(int, int64_t *);

// 給電電圧（WF=0/1: ガウス/微分ガウス）
float vfeed(float t, float delay, float Freq, int WF)
{
	const double tw = 1.27 / Freq;
	const double arg = (t - tw - delay) / (tw / 4);
	const double v = (WF == 0)
		? exp(-arg * arg)
		: sqrt(2) * exp(0.5) * arg * exp(-arg * arg);
	//printf("%e %e %e\n", t, Freq, v);

	return (float)v;
}


// DFT係数
void dftfactor(int Niter, float Freq, float Dt, int WF, float *fcos, float *fsin)
{
	const double pi = 4 * atan(1);
	const double omega = 2 * pi * Freq;

	// 分母
	double sum_r = 0;
	double sum_i = 0;;
	for (int iter = 0; iter <= Niter; iter++) {
		const float t = iter * Dt;
		const float v = vfeed(t, 0, Freq, WF);
		sum_r += cos(omega * t) * v;
		sum_i -= sin(omega * t) * v;
	}
	const float fnorm = (float)((sum_r * sum_r) + (sum_i * sum_i));

	// DFT係数
	for (int iter = 0; iter <= Niter; iter++) {
		const float t = iter * Dt;
		fcos[iter] = (float)cos(omega * t) / fnorm;
		fsin[iter] = (float)sin(omega * t) / fnorm;
	}
}


// PML減衰係数
void pmlfactor(int lpml, float *f)
{
	// PMLパラメーター
	const float pml_m = 2;
	const float pml_r0 = 1e-8f;

	const float c = (float)((pml_m + 1) / (2.0 * lpml) * log(1.0 / pml_r0));
	//printf("%e\n", c);
	/*
	f[0] = f[1] = 0;
	for (int n = 2; n < 2 * lpml + 1; n++) {
		f[n] = (float)(c * pow((n - 1) / (2.0 * lpml), pml_m));
	}
	*/
	f[0] = 0;
	for (int n = 1; n < 2 * lpml + 1; n++) {
		f[n] = (float)(c * pow((n + 0) / (2.0 * lpml), pml_m));
	}
	//for (int n = 0; n < 2 * lpml + 1; n++) printf("%d %f\n", n, f[n]); exit(0);
}


// ランダムに区間を取り出す
// n0, n1 : indexの下限, 上限
// lmin, lmax : 長さの最小, 最大
// i0, i1 : 区間indexの下限, 上限
void getspan(int n0, int n1, int lmin, int lmax, int64_t *seed, int *i0, int *i1)
{
	const float r1 = (float)urand(3, seed);
	const float r2 = (float)urand(3, seed);

	const int ls = (int)(lmin + r1 * (lmax - lmin) + 0.5);  // 長さ
	const int ds = (n1 - n0) - ls;  // 中心の取りうる範囲
	const float pc = n0 + (0.5f * ls) + (r2 * ds);  // 中心

	*i0 = (int)(pc - (0.5 * ls) + 0.5);
	*i1 = (int)(pc + (0.5 * ls) + 0.5);

	assert((*i0 >= n0) && (*i0 <= n1) && (*i1 >= n0) && (*i1 <= n1));
}


// 最小・最大(2D配列)
void getminmax2d(int i0, int i1, int j0, int j1, int db, float **d, float *dmin, float *dmax)
{
	*dmin = +1e12f;
	*dmax = -1e12f;
	for (int i = i0; i < i1; i++) {
	for (int j = j0; j < j1; j++) {
		const float f = d[i][j];
		if (f < *dmin) {
			*dmin = f;
		}
		if (f > *dmax) {
			*dmax = f;
		}
	}
	}

	// dB
	if (db) {
		const float eps = 1e-12f;
		*dmax = 20 * (float)log10(fmax(*dmax, eps));
		*dmin = *dmax - db;
	}
}


// 最小・最大(3D配列)
void getminmax3d(int i0, int i1, int j0, int j1, int k0, int k1, int db, float ***d, float *dmin, float *dmax)
{
	assert((i0 >= 0) && (i1 >= i0) && (j0 >= 0) && (j1 >= j0) && (k0 >= 0) && (k1 >= k0));

	*dmin = +1e12f;
	*dmax = -1e12f;
	for (int i = i0; i < i1; i++) {
	for (int j = j0; j < j1; j++) {
	for (int k = k0; k < k1; k++) {
		const float f = d[i][j][k];
		//printf("%d %d %d %e\n", i, j, k, f);
		if (f < *dmin) {
			*dmin = f;
		}
		if (f > *dmax) {
			*dmax = f;
		}
	}
	}
	}

	// dB
	if (db) {
		const float eps = 1e-12f;
		*dmax = 20 * (float)log10(fmax(*dmax, eps));
		*dmin = *dmax - db;
	}
}


// 最小・最大(4D配列)
void getminmax4d(int i11, int i12, int i21, int i22, int i31, int i32, int i41, int i42, int db, float ****d, float *dmin, float *dmax)
{
	*dmin = +1e12f;
	*dmax = -1e12f;
	for (int i1 = i11; i1 < i12; i1++) {
	for (int i2 = i21; i2 < i22; i2++) {
	for (int i3 = i31; i3 < i32; i3++) {
	for (int i4 = i41; i4 < i42; i4++) {
		const float f = d[i1][i2][i3][i4];
		if (f < *dmin) {
			*dmin = f;
		}
		if (f > *dmax) {
			*dmax = f;
		}
	}
	}
	}
	}

	// dB
	if (db) {
		const float eps = 1e-12f;
		*dmax = 20 * (float)log10(fmax(*dmax, eps));
		*dmin = *dmax - db;
	}
}


// 平均(2D配列)
float average2d(int i0, int i1, int j0, int j1, int abs, float **d)
{
	double sum = 0;
	for (int i = i0; i < i1; i++) {
	for (int j = j0; j < j1; j++) {
		sum += abs ? fabs(d[i][j]) : d[i][j];
	}
	}
	const int num = (i1 - i0) * (j1 - j0);
	if (num > 0) {
		sum /= num;
	}

	return (float)sum;
}


// 平均(3D配列)
float average3d(int i0, int i1, int j0, int j1, int k0, int k1, int abs, float ***d)
{
	double sum = 0;
	for (int i = i0; i < i1; i++) {
	for (int j = j0; j < j1; j++) {
	for (int k = k0; k < k1; k++) {
		sum += abs ? fabs(d[i][j][k]) : d[i][j][k];
	}
	}
	}
	const int num = (i1 - i0) * (j1 - j0) * (k1 - k0);
	if (num > 0) {
		sum /= num;
	}

	return (float)sum;
}


// 平均(4D配列)
float average4d(int i11, int i12, int i21, int i22, int i31, int i32, int i41, int i42, int abs, float ****d)
{
	double sum = 0;
	for (int i1 = i11; i1 < i12; i1++) {
	for (int i2 = i21; i2 < i22; i2++) {
	for (int i3 = i31; i3 < i32; i3++) {
	for (int i4 = i41; i4 < i42; i4++) {
		sum += abs ? fabs(d[i1][i2][i3][i4]) : d[i1][i2][i3][i4];
	}
	}
	}
	}
	const int num = (i12 - i11) * (i22 - i21) * (i32 - i31) * (i42 - i41);
	if (num > 0) {
		sum /= num;
	}

	return (float)sum;
}


// 出力ファイルサイズ[MB]
int file_size(int Ndata, int Nx, int Ny, int Nz, int NTx, int NRx, int Ntime)
{
	size_t size = 0;

	size += (size_t)Ndata * Nx * Ny * Nz * sizeof(unsigned char);  // Epsr or Sigm
	size += (size_t)Ndata * NTx * NRx * Ntime * sizeof(unsigned char);   // Er

	return (int)(size * 1e-6) + 1;

}
