/*
GPR
FDTD (3D)
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#include "../fdtd2d/alloc.h"

// prototypes
extern void make_mesh(int, int, int, float, float, float, int, float *, float *, float *, float *, float *, float *, float *, float *, float *, float *);
extern void txrx(int, int, int, int, int, int, int, int, int **, int, int, int, int ***);
extern void geometry(int, int, int, int64_t, const int [], const float [], const float [], int, int, int, int, float ****, float ****, float, float, float, float, int **);
extern void solve(int, int, int, int, float, float, int, int, int *, int, int **, float *, float *, float *, float *, float *, float *, float *, float *, float *, float ***, float ***, float ****, float ****, float ****, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ****, float ****, float ****, float ****, float ****, float ****, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float **, int, int, float ***, float ***, float ***, float ***, float ***, float ***, float ***);
extern void writedata(int, int, int, int, int, int, int, int, float ****, float ****, const char []);
extern float average4d(int, int, int, int, int, int, int, int, int, float ****);
extern void plot2d_g(int, int, int, int, int, int, float ****, int, int, int, int **, const char []);
extern void plot3d_g(int, int, int, int, int, int, int **, int, int **, const char []);
extern void plot2d_e(int, int, int, int, int, int, int, int, float ****, int, int, int, int **, const char []);
extern void plot2d_r(int, int, int, int, int, float ****, int, int, const char []);
extern int memory_size(int, int, int, int, int, int, int, int, int, int);
extern int file_size(int, int, int, int, int, int, int);
extern void mpi_init(int, char *[], int *, int *);
extern void mpi_close(void);
extern void comm_R(int, int, int, int, int, int, int, float ****);
extern double cputime(void);

const float C = 2.99792458e8f;  // 光速

int main(int argc, char *argv[])
{
	// 定数
	const int   Ndata   = 16;            // データ数
	const int   Nrect   = 2;              // 直方体数
	const int   Lrect[] = {8, 16, 5};     // 直方体パラメーター(最小,最大,余白)（セル数）
	const int   Nsmot   = 1;              // 平滑化回数(0以上)
	const int   Nx      = 80;             // X方向セル数
	const int   Ny      = 30;             // Y方向セル数
	const int   Nz      = 50;             // Z方向セル数
	const int   iTx0    = 10;             // 送信アンテナ位置の開始点（セル数）
	const int   NTx     = 56;             // 送信アンテナ数
	const int   DTx     = 1;              // 送信アンテナ移動間隔（セル数）
	const int   LTxRx   = 5;              // 送受信アンテナ間距離（セル数）
	const int   Lzant   = 5;              // 送信アンテナの地面からの高さ（セル数）
	const int   Lz0     = 10;             // 地上のセル数
	const int   NRx     = 5;              // 受信アンテナ数, 1/3/5,...
	const int   DyRx    = 5;              // 受信アンテナY方向間隔（セル数）
	const float Dx      = 0.025f;         // X方向セルサイズ
	const float Dy      = 0.025f;         // Y方向セルサイズ
	const float Dz      = 0.025f;         // Z方向セルサイズ
	const float Freq    = 600e6f;         // 周波数
	const int   WF      = 0;              // 0:ガウス, 1:微分ガウス
	const int   Labc    = 6;              // PML層数（0のときはMur）
	const int   Ntime   = 224;            // タイムステップ数（初期分除く）

	const float Eps0    = 1.0f;           // 地上の比誘電率
	const float Sig0    = 0.0f;           // 地上の導電率[S/m]
	const float Eps1    = 4.0f;           // 地下の比誘電率
	const float Sig1    = 0.0f;           // 地下の導電率[S/m]
	const float Eps[]   = {1.0f, 1.0f};   // 直方体の誘電率(最小,最大)
	const float Sig[]   = {0.0f, 0.0f};   // 直方体の導電率[S/m](最小,最大)

	const int   Lz = Nz - Lz0;            // 地下のセル数
	const int64_t Seed = 10000;           // 乱数の種, 10000:学習用, 20000:テスト用
	char outfile[BUFSIZ] = "fdtd.bin";    // 出力ファイル名

	// debug
	const int log1 = 1;  // (数値出力) 計算条件とレーダ画像平均(0/1, 通常1)
	const int log2 = 0;  // (数値出力) 収束状況(0/1, 通常0)
	const int fig1 = 0;  // (図形出力) 誘電率分布,送受信点(2D/3D)(0/1, 通常0)
	const int fig2 = 0;  // (図形出力) 電界分布(2D, 複数ページ対応)(0,1,2..., 通常0)
	const int fig3 = 0;  // (図形出力) レーダー画像(2D)(0/1, 通常0)
	const char fig1ev2[] = "fig1.ev2";  // $ ev2d.exe fig1.ev2
	const char fig1ev3[] = "fig1.ev3";  // $ ev3d_otk.exe fig1.ev3
	const char fig2ev2[] = "fig2.ev2";  // $ ev2d.exe fig2.ev2
	const char fig3ev2[] = "fig3.ev2";  // $ ev2d.exe fig3.ev2

	// MPI
	int commsize, commrank;
	mpi_init(argc, argv, &commsize, &commrank);
	//printf("%d %d\n", commsize, commrank);

	// 引数: スレッド数、出力ファイル
	int nthread = 1;
	if (argc > 1) {
		nthread = atoi(argv[1]);
	}
	if (argc > 2) {
		strcpy(outfile, argv[2]);
	}
	//printf("%d %s\n", nthread, outfile);

	// OpenMP
#ifdef _OPENMP
	omp_set_num_threads(nthread);
#endif

	assert((Nx > 0) && (Ny > 0) && (Nz > 0) && (Lz > 0));
	assert((Ntime > 0) && (Labc >= 0));

	// 開始時刻
	const double tstart = cputime();

	// メッシュ作成
	float Dt;
	float *Xn, *Yn, *Zn, *dXn, *dYn, *dZn, *dXc, *dYc, *dZc;
	alloc1d(float, Xn,  Nx + 1)
	alloc1d(float, Yn,  Ny + 1)
	alloc1d(float, Zn,  Nz + 1)
	alloc1d(float, dXn, Nx + 1)
	alloc1d(float, dYn, Ny + 1)
	alloc1d(float, dZn, Nz + 1)
	alloc1d(float, dXc, Nx + 0)
	alloc1d(float, dYc, Ny + 0)
	alloc1d(float, dZc, Nz + 0)
	make_mesh(Nx, Ny, Nz, Dx, Dy, Dz, Lz, &Dt, Xn, Yn, Zn, dXn, dYn, dZn, dXc, dYc, dZc);

	// 送受信点座標
	int **Tx = NULL, ***Rx = NULL;
	alloc2d(int, Tx, NTx, 3)
	alloc3d(int, Rx, NTx, NRx, 3)
	txrx(Nx, Ny, Nz, Lz, Lzant, NTx, iTx0, DTx, Tx, NRx, LTxRx, DyRx, Rx);

	// タイムステップ数 (送信点によらない)
	const float tw = 1.27f / Freq;  // パルス幅
	const float t0 = (float)((2.0 * tw) + (2 * Lzant * Dz * sqrt(Eps0)) / C);  // 波形記録開始時刻
	const int it0 = (int)(t0 / Dt);
	const int Niter = Ntime + it0;
	//printf("%e %e %d %d %d\n", t0, Dt, it0, Ntime, Niter);

	// 計算条件表示
	if (log1 && (commrank == 0)) {
		const int mem = memory_size(nthread, Ndata, Nx, Ny, Nz, NTx, NRx, Niter, Labc, fig2);
		const int filesize = file_size(Ndata, Nx, Ny, Nz, NTx, NRx, Ntime);
		printf("FDTD(3D) data=%d Nx=%d Ny=%d Nz=%d Tx=%d Rx=%d rect=%d(%d) WF=%d ABC=%d\n",
			Ndata, Nx, Ny, Nz, NTx, NRx, Nrect, Nsmot, WF, Labc);
		printf("process=%d thread=%d memory=%dMB iter=%d(%d) dt=%.5fns output=%s(%dMB)\n",
			commsize, nthread, mem, Niter, Ntime, Dt * 1e9, outfile, filesize);
		fflush(stdout);
	}

	// 長方形(debug)
	int **rect = NULL;
	alloc2d(int, rect, Ndata, 6 * Nrect)

	// 誘電率配列(セル中心, xデータ数)
	float ****Epsr, ****Sigm;
	alloc4d(float, Epsr, Ndata, Nx, Ny, Nz)
	alloc4d(float, Sigm, Ndata, Nx, Ny, Nz)

	// 誘電率配列に代入
	geometry(Ndata, Nrect, Nsmot, Seed, Lrect, Eps, Sig, Nx, Ny, Nz, Lz, Epsr, Sigm, Eps0, Sig0, Eps1, Sig1, rect);

	// debug, 誘電率分布図形出力
	if (((fig1 == 1) || (fig1 == -1)) && (commrank == 0)) {
		plot2d_g(1, Nx, Ny, Nz, 5, 10, Epsr, 1, 0, NTx, Tx, fig1ev2);
		plot3d_g(1, Nx, Ny, Nz, Lz, Nrect, rect, NTx, Tx, fig1ev3);
	}
	if (fig1 < 0) {
		exit(0);  // debug: 計算せずに終了する
	}

	// レーダー画像(データ数x送信点数x観測点数xタイムステップ数)
	float ****Vr = NULL;
	alloc4d(float, Vr, Ndata, NTx, NRx, Niter)

	// debug, 全領域調和界, 図形表示用, Yee格子, PML層を含まない, データ0(=スレッド0)のみ
	float ***Ex_r = NULL, ***Ex_i = NULL;
	float ***Ey_r = NULL, ***Ey_i = NULL;
	float ***Ez_r = NULL, ***Ez_i = NULL;
	if (fig2 && (commrank == 0)) {
		alloc3d(float, Ex_r, Nx + 0, Ny + 1, Nz + 1)
		alloc3d(float, Ex_i, Nx + 0, Ny + 1, Nz + 1)
		alloc3d(float, Ey_r, Nx + 1, Ny + 0, Nz + 1)
		alloc3d(float, Ey_i, Nx + 1, Ny + 0, Nz + 1)
		alloc3d(float, Ez_r, Nx + 1, Ny + 1, Nz + 0)
		alloc3d(float, Ez_i, Nx + 1, Ny + 1, Nz + 0)
	}

	// debug, 全領域調和界, 図形表示用, セル中心. PML層を含まない, 送信点の配列, データ0のみ
	// ベクトル合成
	// Y面の集合, YXZに注意
	float ****Ea = NULL;
	if (fig2 && (commrank == 0)) {
		alloc4d(float, Ea, NTx, Ny, Nx, Nz)
	}

	// プロセスごとのデータ番号の範囲（ブロック分割）
	const int ndata = (Ndata + (commsize - 1)) / commsize;
	int idata0 = commrank * ndata;
	if (idata0 > Ndata - 1) {
		idata0 = Ndata - 1;
	}
	int idata1 = idata0 + ndata;
	if (idata1 > Ndata) {
		idata1 = Ndata;
	}
	//printf("%d %d %d %d\n", commsize, commrank, idata0, idata1); fflush(stdout);

	// OpenMP並列領域
#ifdef _OPENMP
#pragma omp parallel
#endif
{
	// Yee格子物性値(スレッド変数, 磁性体は考えないので電界成分のみ)
	float ****PEx = NULL, ****PEy = NULL, ****PEz = NULL;
	alloc4d(float, PEx, Nx + 0 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 1 + 2 * Labc, 2)
	alloc4d(float, PEy, Nx + 1 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 1 + 2 * Labc, 2)
	alloc4d(float, PEz, Nx + 1 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 0 + 2 * Labc, 2)

	// FDTD係数(スレッド変数, 磁性体は考えないので電界成分のみ)
	float ***C1Ex = NULL, ***C2Ex = NULL, ***C1Ey = NULL, ***C2Ey = NULL, ***C1Ez = NULL, ***C2Ez = NULL;
	alloc3d(float, C1Ex, Nx + 0, Ny + 1, Nz + 1)
	alloc3d(float, C2Ex, Nx + 0, Ny + 1, Nz + 1)
	alloc3d(float, C1Ey, Nx + 1, Ny + 0, Nz + 1)
	alloc3d(float, C2Ey, Nx + 1, Ny + 0, Nz + 1)
	alloc3d(float, C1Ez, Nx + 1, Ny + 1, Nz + 0)
	alloc3d(float, C2Ez, Nx + 1, Ny + 1, Nz + 0)

	// ABC変数(スレッド変数)
	float ****Exmury = NULL, ****Exmurz = NULL;
	float ****Eymurz = NULL, ****Eymurx = NULL;
	float ****Ezmurx = NULL, ****Ezmury = NULL;
	float ***Exy = NULL, ***Exz = NULL;
	float ***Eyz = NULL, ***Eyx = NULL;
	float ***Ezx = NULL, ***Ezy = NULL;
	float ***Hxy = NULL, ***Hxz = NULL;
	float ***Hyz = NULL, ***Hyx = NULL;
	float ***Hzx = NULL, ***Hzy = NULL;
	if (Labc == 0) {
		alloc4d(float, Exmury, Nx + 0, Ny + 1, Nz + 1, 2)
		alloc4d(float, Exmurz, Nx + 0, Ny + 1, Nz + 1, 2)
		alloc4d(float, Eymurz, Nx + 1, Ny + 0, Nz + 1, 2)
		alloc4d(float, Eymurx, Nx + 1, Ny + 0, Nz + 1, 2)
		alloc4d(float, Ezmurx, Nx + 1, Ny + 1, Nz + 0, 2)
		alloc4d(float, Ezmury, Nx + 1, Ny + 1, Nz + 0, 2)
	}
	else {
		alloc3d(float, Exy, Nx + 0 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 1 + 2 * Labc)
		alloc3d(float, Exz, Nx + 0 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 1 + 2 * Labc)
		alloc3d(float, Eyz, Nx + 1 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 1 + 2 * Labc)
		alloc3d(float, Eyx, Nx + 1 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 1 + 2 * Labc)
		alloc3d(float, Ezx, Nx + 1 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 0 + 2 * Labc)
		alloc3d(float, Ezy, Nx + 1 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 0 + 2 * Labc)
		alloc3d(float, Hxy, Nx + 1 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 0 + 2 * Labc)
		alloc3d(float, Hxz, Nx + 1 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 0 + 2 * Labc)
		alloc3d(float, Hyz, Nx + 0 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 0 + 2 * Labc)
		alloc3d(float, Hyx, Nx + 0 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 0 + 2 * Labc)
		alloc3d(float, Hzx, Nx + 0 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 1 + 2 * Labc)
		alloc3d(float, Hzy, Nx + 0 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 1 + 2 * Labc)
	}

	// 電磁界配列(スレッド変数)
	float ***Ex, ***Ey, ***Ez, ***Hx, ***Hy, ***Hz;
	alloc3d(float, Ex, Nx + 0 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 1 + 2 * Labc)
	alloc3d(float, Ey, Nx + 1 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 1 + 2 * Labc)
	alloc3d(float, Ez, Nx + 1 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 0 + 2 * Labc)
	alloc3d(float, Hx, Nx + 1 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 0 + 2 * Labc)
	alloc3d(float, Hy, Nx + 0 + 2 * Labc, Ny + 1 + 2 * Labc, Nz + 0 + 2 * Labc)
	alloc3d(float, Hz, Nx + 0 + 2 * Labc, Ny + 0 + 2 * Labc, Nz + 1 + 2 * Labc)

	// スレッド番号
	int ithread = 0;
#ifdef _OPENMP
	ithread = omp_get_thread_num();
#endif
	// データに関するループ
	//for (int idata = ithread; idata < Ndata; idata += nthread) {
	for (int idata = idata0 + ithread; idata < idata1; idata += nthread) {
		//printf("%d %d %d\n", commrank, ithread, idata); fflush(stdout);
		const int ilog = log2 && (commrank == 0) && (idata == 0);  // 収束状況(データ0のみ)
		const int iout = fig2 && (commrank == 0) && (idata == 0);  // 電界分布(データ0のみ)
		if (log1 && (commrank == 0)) {printf("%d %d\n", idata, ithread); fflush(stdout);}  // 経過出力

		// 送信点に関するループ
		for (int itx = 0; itx < NTx; itx++) {
			solve(
				Niter, Nx, Ny, Nz, Freq, Dt, Labc, WF,
				Tx[itx], NRx, Rx[itx],
				Xn, Yn, Zn, dXn, dYn, dZn, dXc, dYc, dZc,
				Epsr[idata], Sigm[idata], PEx, PEy, PEz,
				Ex, Ey, Ez, Hx, Hy, Hz,
				C1Ex, C2Ex, C1Ey, C2Ey, C1Ez, C2Ez,
				Exmury, Exmurz, Eymurz, Eymurx, Ezmurx, Ezmury,
				Exy, Exz, Eyz, Eyx, Ezx, Ezy,
				Hxy, Hxz, Hyz, Hyx, Hzx, Hzy,
				Vr[idata][itx],
				ilog, iout, Ex_r, Ex_i, Ey_r, Ey_i, Ez_r, Ez_i, (iout ? Ea[itx] : NULL));
		}
	}
}

	// レーダー画像をrootに集める
	if (commsize > 1) {
		comm_R(commsize, commrank, idata0, idata1, NTx, NRx, Niter, Vr);
	}

	// 出力:root
	if (commrank == 0) {
		// ファイル出力
		// ファイルサイズを小さくするために誘電率と導電率のどちらか一つを選択する
		writedata(Ndata, Nx, Ny, Lz, NTx, NRx, Niter, it0, Epsr, Vr, outfile);  // 誘電率
		//writedata(Ndata, Nx, Ny, Lz, NTx, NRx, Niter, it0, Sigm, Vr, outfile);  // 導電率

		// debug, レーダー画像平均数値出力
		if (log1) {
			const float rave = average4d(0, Ndata, 0, NTx, 0, NRx, 0, Niter, 1, Vr);
			printf("Rave=%.8e\n", rave);
		}

		// debug, 電界分布図形出力
		if (fig2) {
			plot2d_e(1, 1, NTx, 1, Nx, Ny, Nz, Lz, Ea, 100, 0, Nrect, rect, fig2ev2);
		}

		// debug, レーダー画像図形出力
		if (fig3) {
			plot2d_r(Ndata, 1, NTx, NRx, Niter, Vr, it0, 2, fig3ev2);
		}
	}

	// 終了時刻
	const double tstop = cputime();
	if (log1 && (commrank == 0)) {
		printf("CPU=%.3f[sec]\n", tstop - tstart);
		fflush(stdout);
	}

	// MPI
	mpi_close();

	return 0;
}
