/*
comm.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <assert.h>

#ifdef _MPI
#include <mpi.h>
#endif

// init
void mpi_init(int argc, char *argv[], int *commsize, int *commrank)
{
#ifdef _MPI
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, commsize);
	MPI_Comm_rank(MPI_COMM_WORLD, commrank);
#else
	argc = argc;
	argv = argv;
	*commsize = 1;
	*commrank = 0;
#endif
}


// close
void mpi_close(void)
{
#ifdef _MPI
	MPI_Finalize();
#endif
}


// レーダー画像をrootに集める
void comm_R(int commsize, int commrank, int idata0, int idata1, int NTx, int NRx, int Niter, float ****r)
{
#ifdef _MPI
	MPI_Barrier(MPI_COMM_WORLD);  // 必須ではないが念のため

	if (commrank > 0) {
		// 送信(root以外)
		int i_sendbuf[2];
		i_sendbuf[0] = idata0;
		i_sendbuf[1] = idata1;
		MPI_Send(i_sendbuf, 2, MPI_INT, 0, 0, MPI_COMM_WORLD);
		//printf("%d %d %d %d\n", commsize, commrank, i_sendbuf[0], i_sendbuf[1]); fflush(stdout);
		const int count = (idata1 - idata0) * NTx * NRx * Niter;
		float *f_sendbuf = (float *)malloc(count * sizeof(float));
		//printf("%d %d %d\n", commsize, commrank, count); fflush(stdout);
		int num = 0;
		for (int idata = idata0; idata < idata1; idata++) {
		for (int itx = 0; itx < NTx; itx++) {
		for (int irx = 0; irx < NRx; irx++) {
		for (int it = 0; it < Niter; it++) {
			f_sendbuf[num++] = r[idata][itx][irx][it];
		}
		}
		}
		}
		assert(num == count);
		MPI_Send(f_sendbuf, count, MPI_FLOAT, 0, 0, MPI_COMM_WORLD);
		free(f_sendbuf);
	}
	else {
		// 受信(root)
		MPI_Status status;
		int i_recvbuf[2];
		for (int rank = 1; rank < commsize; rank++) {
			MPI_Recv(i_recvbuf, 2, MPI_INT, rank, 0, MPI_COMM_WORLD, &status);
			const int i_data0 = i_recvbuf[0];
			const int i_data1 = i_recvbuf[1];
			//printf("%d %d %d %d\n", commsize, commrank, i_data0, i_data1); fflush(stdout);
			const int count = (i_data1 - i_data0) * NTx * NRx * Niter;
			float *f_recvbuf = (float *)malloc(count * sizeof(float));
			//printf("%d %d %d\n", commsize, commrank, count); fflush(stdout);
			MPI_Recv(f_recvbuf, count, MPI_FLOAT, rank, 0, MPI_COMM_WORLD, &status);
			int num = 0;
			for (int idata = i_data0; idata < i_data1; idata++) {
			for (int itx = 0; itx < NTx; itx++) {
			for (int irx = 0; irx < NRx; irx++) {
			for (int it = 0; it < Niter; it++) {
				r[idata][itx][irx][it] = f_recvbuf[num++];
			}
			}
			}
			}
			assert(num == count);
			free(f_recvbuf);
		}
	}
#else
	// dummy
	commsize = commsize;
	commrank = commrank;
	idata0 = idata0;
	idata1 = idata1;
	NTx = NTx;
	NRx = NRx;
	Niter = Niter;
	r = r;
#endif
}


// CPU時間
double cputime(void)
{
#ifdef _MPI
	MPI_Barrier(MPI_COMM_WORLD);
	return MPI_Wtime();
#else
#ifdef _WIN32
	return (double)clock() / CLOCKS_PER_SEC;
#else
	struct timespec ts;
	clock_gettime(CLOCK_REALTIME, &ts);
	return ts.tv_sec + (ts.tv_nsec * 1e-9);
#endif  // _WIN32
#endif  // _MPI
}
